import abc
from typing import Union, Generic, TypeVar, Optional, overload

from prompt_toolkit.layout import Layout
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Attrs, Style, StyleTransformation

DT = TypeVar("DT")
RT = TypeVar("RT")


NO_ANSWER = ...


class UndefinedType:
    def __str__(self):
        return "Undefined"

    def __bool__(self):
        return False


UNDEFINED = UndefinedType()


class CancelledError(Exception):
    """User cancelled answer."""


class DisableColorTransformation(StyleTransformation):
    def __init__(self, no_ansi: bool = False):
        self.no_ansi = no_ansi

    def transform_attrs(self, attrs: Attrs) -> Attrs:
        if self.no_ansi:
            return Attrs("", "", False, False, False, False, False, False, False)
        return attrs


class BasePrompt(abc.ABC, Generic[RT]):
    @abc.abstractmethod
    def _build_layout(self) -> Layout:
        raise NotImplementedError

    @abc.abstractmethod
    def _build_style(self, style: Style) -> Style:
        raise NotImplementedError

    @abc.abstractmethod
    def _build_keybindings(self) -> KeyBindings:
        raise NotImplementedError

    def _build_application(self, no_ansi: bool, style: Style) -> Application:
        return Application(
            layout=self._build_layout(),
            style=self._build_style(style),
            style_transformation=DisableColorTransformation(no_ansi),
            key_bindings=self._build_keybindings(),
            mouse_support=True,
        )

    @overload
    def prompt(
        self,
        default: UndefinedType = UNDEFINED,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> RT:
        ...

    @overload
    def prompt(
        self,
        default: DT,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> Union[DT, RT]:
        ...

    def prompt(
        self,
        default: Union[DT, UndefinedType] = UNDEFINED,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> Union[DT, RT]:
        app = self._build_application(no_ansi=no_ansi, style=style or Style([]))
        result: RT = app.run()
        if result is not NO_ANSWER:
            return result
        if default is UNDEFINED:
            raise CancelledError("No answer selected!")
        return default  # type: ignore

    @overload
    async def prompt_async(
        self,
        default: UndefinedType = UNDEFINED,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> RT:
        ...

    @overload
    async def prompt_async(
        self,
        default: DT,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> Union[DT, RT]:
        ...

    async def prompt_async(
        self,
        default: Union[DT, UndefinedType] = UNDEFINED,
        no_ansi: bool = False,
        style: Optional[Style] = None,
    ) -> Union[DT, RT]:
        app = self._build_application(no_ansi=no_ansi, style=style or Style([]))
        result: RT = await app.run_async()
        if result is not NO_ANSWER:
            return result
        if default is UNDEFINED:
            raise CancelledError("No answer selected!")
        return default  # type: ignore



from typing import List, Optional

from noneprompt.utils import build_style
from noneprompt.prompts import Choice, ListPrompt

from . import subparsers
from ._options import (
    pointer_option,
    annotation_option,
    answer_style_option,
    filter_style_option,
    select_style_option,
    pointer_style_option,
    question_mark_option,
    question_style_option,
    unselect_style_option,
    annotation_style_option,
    question_mark_style_option,
)


def list_prompt_main(
    question: str,
    choices: List[str],
    default_select: Optional[int] = None,
    *,
    no_ansi: bool = False,
    default: Optional[str] = None,
    question_style: Optional[str] = None,
    question_mark: Optional[str] = None,
    question_mark_style: Optional[str] = None,
    annotation: Optional[str] = None,
    annotation_style: Optional[str] = None,
    filter_style: Optional[str] = None,
    answer_style: Optional[str] = None,
    pointer: Optional[str] = None,
    pointer_style: Optional[str] = None,
    select_style: Optional[str] = None,
    unselect_style: Optional[str] = None,
    **kwargs,
):
    prompt = ListPrompt(
        question,
        choices=[Choice(c) for c in choices],
        default_select=default_select,
        question_mark=question_mark,
        pointer=pointer,
        annotation=annotation,
    )
    prompt.prompt(
        default=default,
        no_ansi=no_ansi,
        style=build_style(
            question=question_style,
            questionmark=question_mark_style,
            annotation=annotation_style,
            filter=filter_style,
            answer=answer_style,
            pointer=pointer_style,
            selected=select_style,
            unselected=unselect_style,
        ),
    )


list_prompt = subparsers.add_parser(
    "list",
    description="List prompt.",
    parents=[
        question_mark_option,
        question_mark_style_option,
        question_style_option,
        annotation_option,
        annotation_style_option,
        filter_style_option,
        answer_style_option,
        pointer_option,
        pointer_style_option,
        select_style_option,
        unselect_style_option,
    ],
)
list_prompt.add_argument(
    "-d", "--default-select", type=int, required=False, help="default select"
)
list_prompt.add_argument("question", help="prompt question")
list_prompt.add_argument("choices", nargs="+", help="choices")
list_prompt.set_defaults(func=list_prompt_main)



from typing import Optional

from prompt_toolkit.styles import Style
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout import Layout
from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.application import get_app
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.validation import Validator
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent

from noneprompt.utils import BOOLEAN_STRING, str2bool

from ._base import NO_ANSWER, BasePrompt


class ConfirmPrompt(BasePrompt[bool]):
    """Simple Confirm Prompt.

    Style class guide:

    ```
    [?] Choose a choice and return? (Y/n)
    └┬┘ └──────────────┬──────────┘ └─┬─┘
    questionmark    question      annotation
    ```
    """

    def __init__(
        self,
        question: str,
        default_choice: Optional[bool] = None,
        *,
        question_mark: Optional[str] = None,
    ):
        self.question: str = question
        self.question_mark: str = "[?]" if question_mark is None else question_mark
        self.default_choice: Optional[bool] = default_choice

    def _reset(self):
        self._answered: bool = False
        self._buffer: Buffer = Buffer(
            validator=Validator.from_callable(self._validate),
            name=DEFAULT_BUFFER,
            accept_handler=self._submit,
        )

    def _build_layout(self) -> Layout:
        self._reset()
        return Layout(
            HSplit(
                [
                    Window(
                        BufferControl(self._buffer, lexer=SimpleLexer("class:answer")),
                        dont_extend_height=True,
                        get_line_prefix=self._get_prompt,
                    )
                ]
            )
        )

    def _build_style(self, style: Style) -> Style:
        default = Style(
            [
                ("questionmark", "fg:#673AB7 bold"),
                ("question", "bold"),
                ("annotation", "fg:#7F8C8D"),
                ("answer", "fg:#FF9D00"),
            ]
        )
        return Style([*default.style_rules, *style.style_rules])

    def _build_keybindings(self) -> KeyBindings:
        kb = KeyBindings()

        @kb.add("enter", eager=True)
        def enter(event: KeyPressEvent):
            self._buffer.validate_and_handle()

        @kb.add("c-c", eager=True)
        @kb.add("c-q", eager=True)
        def quit(event: KeyPressEvent):
            event.app.exit(result=NO_ANSWER)

        return kb

    def _get_prompt(self, line_number: int, wrap_count: int) -> AnyFormattedText:
        prompts: AnyFormattedText = []
        if self.question_mark:
            prompts.append(("class:questionmark", self.question_mark))
            prompts.append(("", " "))
        prompts.append(("class:question", self.question.strip()))
        prompts.append(("", " "))
        if not self._answered:
            if self.default_choice is None:
                prompts.append(("class:annotation", "(y/n)"))
            elif self.default_choice:
                prompts.append(("class:annotation", "(Y/n)"))
            else:
                prompts.append(("class:annotation", "(y/N)"))
            prompts.append(("", " "))
        return prompts

    def _validate(self, input: str) -> bool:
        if not input and self.default_choice is None:
            return False
        elif input and input.lower() not in BOOLEAN_STRING:
            return False
        return True

    def _submit(self, buffer: Buffer) -> bool:
        self._answered = True
        input = buffer.document.text
        if not input:
            buffer.insert_text("Yes" if self.default_choice else "No")
            get_app().exit(result=self.default_choice)
        else:
            get_app().exit(result=str2bool(input))
        return True



from dataclasses import dataclass
from typing import Generic, TypeVar

RT = TypeVar("RT")


@dataclass
class Choice(Generic[RT]):
    """Choice item for sequence like prompts."""

    name: str
    data: RT = None  # type: ignore



from argparse import ArgumentParser

# question mark option
question_mark_option = ArgumentParser(add_help=False)
question_mark_option.add_argument(
    "-qm", "--question-mark", required=False, help="question mark"
)

# question mark style option
question_mark_style_option = ArgumentParser(add_help=False)
question_mark_style_option.add_argument(
    "-qms", "--question-mark-style", required=False, help="question mark style"
)

# question style option
question_style_option = ArgumentParser(add_help=False)
question_style_option.add_argument(
    "-qs", "--question-style", required=False, help="question style"
)

# annotation option
annotation_option = ArgumentParser(add_help=False)
annotation_option.add_argument(
    "-an", "--annotation", required=False, help="annotation texts"
)

# annotation style option
annotation_style_option = ArgumentParser(add_help=False)
annotation_style_option.add_argument(
    "-ans", "--annotation-style", required=False, help="annotation style"
)

# input style option
input_style_option = ArgumentParser(add_help=False)
input_style_option.add_argument(
    "-is", "--input-style", required=False, help="input style"
)

# answer style option
answer_style_option = ArgumentParser(add_help=False)
answer_style_option.add_argument(
    "-as", "--answer-style", required=False, help="answer style"
)

# pointer option
pointer_option = ArgumentParser(add_help=False)
pointer_option.add_argument("-p", "--pointer", required=False, help="pointer mark")

# pointer style option
pointer_style_option = ArgumentParser(add_help=False)
pointer_style_option.add_argument(
    "-ps", "--pointer-style", required=False, help="pointer mark style"
)

# sign option
sign_option = ArgumentParser(add_help=False)
sign_option.add_argument("-s", "--sign", required=False, help="sign mark")

# sign style option
sign_style_option = ArgumentParser(add_help=False)
sign_style_option.add_argument(
    "-ss", "--sign-style", required=False, help="sign mark style"
)

# unsign option
unsign_option = ArgumentParser(add_help=False)
unsign_option.add_argument("-u", "--unsign", required=False, help="unsign mark")

# unsign style option
unsign_style_option = ArgumentParser(add_help=False)
unsign_style_option.add_argument(
    "-us", "--unsign-style", required=False, help="unsign mark style"
)

# select style option
select_style_option = ArgumentParser(add_help=False)
select_style_option.add_argument(
    "-ses", "--select-style", required=False, help="select mark style"
)

# unselect style option
unselect_style_option = ArgumentParser(add_help=False)
unselect_style_option.add_argument(
    "-unses", "--unselect-style", required=False, help="unselect mark style"
)

# filter style option
filter_style_option = ArgumentParser(add_help=False)
filter_style_option.add_argument("-f", "--filter", required=False, help="filter style")



from ._choice import Choice as Choice
from .list import ListPrompt as ListPrompt
from ._base import BasePrompt as BasePrompt
from .input import InputPrompt as InputPrompt
from ._base import CancelledError as CancelledError
from .confirm import ConfirmPrompt as ConfirmPrompt
from .checkbox import CheckboxPrompt as CheckboxPrompt



from typing import Union, Callable, Optional

from prompt_toolkit.styles import Style
from prompt_toolkit.layout import Layout
from prompt_toolkit.application import get_app
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.validation import Validator
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.lexers import SimpleLexer, DynamicLexer
from prompt_toolkit.filters import Filter, Condition, is_done
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.processors import PasswordProcessor, ConditionalProcessor
from prompt_toolkit.layout.containers import (
    Float,
    HSplit,
    Window,
    FloatContainer,
    ConditionalContainer,
)

from ._base import NO_ANSWER, BasePrompt


class InputPrompt(BasePrompt[str]):
    """Simple Input Prompt.

    Style class guide:

    ```
    [?] Choose a choice and return? answer
    └┬┘ └──────────────┬──────────┘ └──┬─┘
    questionmark    question      input/answer

    Invalid input
    └─────┬─────┘
        error
    ```
    """

    def __init__(
        self,
        question: str,
        default_text: Optional[str] = None,
        is_password: Union[bool, Filter] = False,
        *,
        question_mark: Optional[str] = None,
        validator: Optional[Callable[[str], bool]] = None,
        error_message: Optional[str] = "Invalid input",
    ):
        self.question: str = question
        self.default_text: Optional[str] = default_text
        self.is_password: Union[bool, Filter] = is_password
        self.question_mark: str = "[?]" if question_mark is None else question_mark
        self.validator: Optional[Callable[[str], bool]] = validator
        self.error_message: Optional[str] = error_message

    def _reset(self):
        self._answered: bool = False
        self._invalid: bool = False
        self._buffer: Buffer = Buffer(
            name=DEFAULT_BUFFER,
            validator=self.validator
            and Validator.from_callable(self.validator, move_cursor_to_end=True),
            accept_handler=self._submit,
            multiline=False,
            on_text_changed=self._on_text_changed,
        )
        if self.default_text:
            self._buffer.insert_text(self.default_text)

    def _build_layout(self) -> Layout:
        self._reset()
        return Layout(
            HSplit(
                [
                    Window(
                        BufferControl(
                            self._buffer,
                            input_processors=[
                                ConditionalProcessor(
                                    PasswordProcessor(), self.is_password
                                )
                            ],
                            lexer=DynamicLexer(
                                lambda: (
                                    SimpleLexer("class:answer")
                                    if self._answered
                                    else SimpleLexer("class:input")
                                )
                            ),
                        ),
                        dont_extend_height=True,
                        get_line_prefix=self._get_prompt,
                    ),
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(self._get_error_prompt),
                            height=1,
                            dont_extend_height=True,
                            always_hide_cursor=True,
                        ),
                        Condition(
                            lambda: self.error_message is not None and self._invalid
                        )
                        & ~is_done,
                    ),
                ]
            )
        )

    def _build_style(self, style: Style) -> Style:
        default = Style(
            [
                ("questionmark", "fg:#673AB7 bold"),
                ("question", "bold"),
                ("answer", "fg:#FF9D00"),
                ("error", "bg:#FF0000"),
            ]
        )
        return Style([*default.style_rules, *style.style_rules])

    def _build_keybindings(self) -> KeyBindings:
        kb = KeyBindings()

        @kb.add("enter", eager=True)
        def enter(event: KeyPressEvent):
            self._buffer.validate_and_handle()
            if self._buffer.validation_state == ValidationState.INVALID:
                self._invalid = True

        @kb.add("c-c", eager=True)
        @kb.add("c-q", eager=True)
        def quit(event: KeyPressEvent):
            event.app.exit(result=NO_ANSWER)

        return kb

    def _get_prompt(self, line_number: int, wrap_count: int) -> AnyFormattedText:
        prompts: AnyFormattedText = []
        if self.question_mark:
            prompts.append(("class:questionmark", self.question_mark))
            prompts.append(("", " "))
        prompts.append(("class:question", self.question.strip()))
        prompts.append(("", " "))
        return prompts

    def _get_error_prompt(self) -> AnyFormattedText:
        return self.error_message and [("class:error", self.error_message)]

    def _submit(self, buffer: Buffer) -> bool:
        self._answered = True
        get_app().exit(result=buffer.document.text)
        return True

    def _on_text_changed(self, buffer: Buffer) -> None:
        # reset validation state
        if self._invalid == True:
            self._invalid = False



from typing import Optional

from noneprompt.utils import build_style
from noneprompt.prompts import InputPrompt

from . import subparsers
from ._options import (
    input_style_option,
    answer_style_option,
    question_mark_option,
    question_style_option,
    question_mark_style_option,
)


def input_prompt_main(
    question: str,
    default_text: Optional[str] = None,
    *,
    no_ansi: bool = False,
    default: Optional[str] = None,
    question_style: Optional[str] = None,
    question_mark: Optional[str] = None,
    question_mark_style: Optional[str] = None,
    input_style: Optional[str] = None,
    answer_style: Optional[str] = None,
    **kwargs,
):
    prompt = InputPrompt(
        question, default_text=default_text, question_mark=question_mark
    )
    prompt.prompt(
        default=default,
        no_ansi=no_ansi,
        style=build_style(
            question=question_style,
            questionmark=question_mark_style,
            input=input_style,
            answer=answer_style,
        ),
    )


input_prompt = subparsers.add_parser(
    "input",
    description="Input prompt.",
    parents=[
        question_mark_option,
        question_mark_style_option,
        question_style_option,
        input_style_option,
        answer_style_option,
    ],
)
input_prompt.add_argument(
    "-d",
    "--default-text",
    required=False,
    help="default text",
)
input_prompt.add_argument("question", help="prompt question")
input_prompt.set_defaults(func=input_prompt_main)



from .prompts import Choice as Choice
from .prompts import BasePrompt as BasePrompt
from .prompts import ListPrompt as ListPrompt
from .prompts import InputPrompt as InputPrompt
from .prompts import ConfirmPrompt as ConfirmPrompt
from .prompts import CancelledError as CancelledError
from .prompts import CheckboxPrompt as CheckboxPrompt



from typing import Dict, Callable
from argparse import ArgumentParser

parser = ArgumentParser(prog="noneprompt", description="A simple prompt toolkit.")
parser.add_argument(
    "--no-ansi", action="store_true", required=False, help="disable ANSI colors"
)
parser.add_argument(
    "-d", "--default", required=False, help="default answer when cancelled"
)

subparsers = parser.add_subparsers(title="prompt type", required=True)

from . import list, input, confirm, checkbox


def main():
    result = parser.parse_args()
    result = vars(result)
    prompt: Callable[..., None] = result.pop("func")
    prompt(**result)



from typing import List, Optional

from noneprompt.utils import build_style
from noneprompt.prompts import Choice, CheckboxPrompt

from . import subparsers
from ._options import (
    sign_option,
    unsign_option,
    pointer_option,
    annotation_option,
    sign_style_option,
    answer_style_option,
    select_style_option,
    unsign_style_option,
    pointer_style_option,
    question_mark_option,
    question_style_option,
    unselect_style_option,
    annotation_style_option,
    question_mark_style_option,
)


def checkbox_prompt_main(
    question: str,
    choices: List[str],
    default_select: Optional[List[int]] = None,
    *,
    no_ansi: bool = False,
    default: Optional[str] = None,
    question_style: Optional[str] = None,
    question_mark: Optional[str] = None,
    question_mark_style: Optional[str] = None,
    annotation: Optional[str] = None,
    annotation_style: Optional[str] = None,
    answer_style: Optional[str] = None,
    pointer: Optional[str] = None,
    pointer_style: Optional[str] = None,
    sign: Optional[str] = None,
    sign_style: Optional[str] = None,
    unsign: Optional[str] = None,
    unsign_style: Optional[str] = None,
    select_style: Optional[str] = None,
    unselect_style: Optional[str] = None,
    **kwargs,
):
    prompt = CheckboxPrompt(
        question,
        choices=[Choice(c) for c in choices],
        default_select=default_select,
        question_mark=question_mark,
        pointer=pointer,
        selected_sign=sign,
        unselected_sign=unsign,
        annotation=annotation,
    )
    prompt.prompt(
        default=default,
        no_ansi=no_ansi,
        style=build_style(
            question=question_style,
            questionmark=question_mark_style,
            annotation=annotation_style,
            answer=answer_style,
            pointer=pointer_style,
            sign=sign_style,
            unsign=unsign_style,
            selected=select_style,
            unselected=unselect_style,
        ),
    )


checkbox_prompt = subparsers.add_parser(
    "checkbox",
    description="Checkbox prompt.",
    parents=[
        question_mark_option,
        question_mark_style_option,
        question_style_option,
        annotation_option,
        annotation_style_option,
        answer_style_option,
        pointer_option,
        pointer_style_option,
        sign_option,
        sign_style_option,
        unsign_option,
        unsign_style_option,
        select_style_option,
        unselect_style_option,
    ],
)
checkbox_prompt.add_argument(
    "-d",
    "--default-select",
    action="append",
    type=int,
    required=False,
    help="default select",
)
checkbox_prompt.add_argument("question", help="prompt question")
checkbox_prompt.add_argument("choices", nargs="+", help="choices")
checkbox_prompt.set_defaults(func=checkbox_prompt_main)



from typing import Optional

from prompt_toolkit.styles import Style

BOOLEAN_STRING_TRUE = {"yes", "y", "true", "t", "1"}
BOOLEAN_STRING_FALSE = {"no", "n", "false", "f", "0"}
BOOLEAN_STRING = BOOLEAN_STRING_TRUE | BOOLEAN_STRING_FALSE


def str2bool(value: str):
    return value.lower() in BOOLEAN_STRING_TRUE


def build_style(**kwargs: Optional[str]) -> Style:
    return Style([(k, v) for k, v in kwargs.items() if v is not None])



import os
from time import time
from functools import lru_cache
from typing import List, TypeVar, Callable, Optional

from prompt_toolkit.styles import Style
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout import Layout
from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.application import get_app
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Condition, is_done
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.containers import (
    Float,
    HSplit,
    Window,
    FloatContainer,
    ConditionalContainer,
)

from ._choice import Choice
from ._base import NO_ANSWER, BasePrompt

RT = TypeVar("RT")


class ListPrompt(BasePrompt[Choice[RT]]):
    """RadioList Prompt that supports auto scrolling.

    Style class guide:

    ```
    [?] Choose a choice and return? (Use ↑ and ↓ to choose, Enter to submit) input
    └┬┘ └──────────────┬──────────┘ └────────────────────┬─────────────────┘ └─┬─┘
    questionmark    question                         annotation              filter

     ❯  choice selected
    └┬┘ └───────┬─────┘
    pointer  selected

        choice unselected
        └───────┬───────┘
            unselected

    Invalid input
    └─────┬─────┘
        error
    ```
    """

    def __init__(
        self,
        question: str,
        choices: List[Choice[RT]],
        allow_filter: bool = True,
        default_select: Optional[int] = None,
        *,
        question_mark: Optional[str] = None,
        pointer: Optional[str] = None,
        annotation: Optional[str] = None,
        max_height: Optional[int] = None,
        validator: Optional[Callable[[Choice[RT]], bool]] = None,
        error_message: Optional[str] = "Invalid selection",
    ):
        self.question: str = question
        self.choices: List[Choice[RT]] = choices
        self.allow_filter: bool = allow_filter
        self.question_mark: str = "[?]" if question_mark is None else question_mark
        self.pointer: str = "❯" if pointer is None else pointer
        self.annotation: str = (
            "(Use ↑ and ↓ to choose, Enter to submit)"
            if annotation is None
            else annotation
        )
        self.validator: Optional[Callable[[Choice[RT]], bool]] = validator
        self.error_message: Optional[str] = error_message

        self._index: int = (default_select or 0) % len(self.choices)
        self._display_index: int = 0
        self._max_height: Optional[int] = max_height

        self._last_mouse_up: float = 0
        self._last_mouse_index: int = 0

    @property
    def max_height(self) -> int:
        return self._max_height or os.get_terminal_size().lines

    @property
    def filtered_choices(self) -> List[Choice[RT]]:
        return [
            choice
            for choice in self.choices
            if self._buffer.text.lower() in choice.name.lower()
        ]

    def _reset(self):
        self._invalid: bool = False
        self._answered: Optional[Choice] = None
        self._reset_choice_layout()
        self._buffer: Buffer = Buffer(
            name=DEFAULT_BUFFER,
            multiline=False,
            on_text_changed=lambda b: self._reset_choice_layout(),
        )

    def _reset_choice_layout(self):
        self._index: int = 0
        self._display_index: int = 0
        self._reset_error()

    def _reset_error(self):
        self._invalid = False

    def _build_layout(self) -> Layout:
        self._reset()
        if self.allow_filter:
            prompt_line = Window(
                BufferControl(
                    self._buffer,
                    lexer=SimpleLexer("class:filter"),
                    focus_on_click=~is_done,
                ),
                dont_extend_height=True,
                get_line_prefix=self._get_line_prefix,
            )
        else:
            prompt_line = Window(
                FormattedTextControl(self._get_prompt),
                height=Dimension(1),
                dont_extend_height=True,
                always_hide_cursor=True,
            )

        return Layout(
            FloatContainer(
                HSplit(
                    [
                        prompt_line,
                        ConditionalContainer(
                            Window(
                                FormattedTextControl(self._get_choices_prompt),
                                height=Dimension(1),
                                dont_extend_height=True,
                                always_hide_cursor=True,
                            ),
                            ~is_done,
                        ),
                    ]
                ),
                [
                    Float(
                        ConditionalContainer(
                            Window(
                                FormattedTextControl(self._get_error_prompt),
                                height=1,
                                dont_extend_height=True,
                                always_hide_cursor=True,
                            ),
                            Condition(
                                lambda: self.error_message is not None and self._invalid
                            )
                            & ~is_done,
                        ),
                        bottom=0,
                        left=0,
                    )
                ],
            )
        )

    def _build_style(self, style: Style) -> Style:
        default = Style(
            [
                ("questionmark", "fg:#673AB7 bold"),
                ("question", "bold"),
                ("answer", "fg:#FF9D00"),
                ("annotation", "fg:#7F8C8D"),
                ("selected", "fg:ansigreen noreverse"),
                ("error", "bg:#FF0000"),
            ]
        )
        return Style([*default.style_rules, *style.style_rules])

    def _build_keybindings(self) -> KeyBindings:
        kb = KeyBindings()

        @kb.add("up", eager=True)
        def previous(event: KeyPressEvent):
            self._reset_error()
            self._handle_up()

        @kb.add("down", eager=True)
        def next(event: KeyPressEvent):
            self._reset_error()
            self._handle_down()

        @kb.add("enter", eager=True)
        def enter(event: KeyPressEvent):
            self._reset_error()
            self._finish()

        @kb.add("c-c", eager=True)
        @kb.add("c-q", eager=True)
        def quit(event: KeyPressEvent):
            self._reset_error()
            event.app.exit(result=NO_ANSWER)

        return kb

    def _finish(self) -> None:
        # no answer selected
        if not self.filtered_choices:
            return

        # get result first
        result = self.filtered_choices[self._index]
        # validate result
        if self.validator is not None:
            self._invalid = not self.validator(result)
            if self._invalid:
                return

        self._answered = result
        # then clear buffer
        self._buffer.reset()
        get_app().exit(result=result)

    def _get_line_prefix(self, line_number: int, wrap_count: int) -> AnyFormattedText:
        return self._get_prompt()

    def _get_prompt(self) -> AnyFormattedText:
        prompts: AnyFormattedText = []
        if self.question_mark:
            prompts.append(("class:questionmark", self.question_mark))
            prompts.append(("", " "))
        prompts.append(("class:question", self.question.strip()))
        prompts.append(("", " "))
        if self._answered:
            prompts.append(("class:answer", self._answered.name.strip()))
        else:
            prompts.append(("class:annotation", self.annotation))
            prompts.append(("", " "))
        return prompts

    def _get_choices_prompt(self) -> AnyFormattedText:
        max_num = self.max_height - 1

        prompts: AnyFormattedText = []
        for index, choice in enumerate(
            self.filtered_choices[self._display_index : self._display_index + max_num]
        ):
            current_index = index + self._display_index
            if current_index == self._index:
                prompts.append(
                    (
                        "class:pointer",
                        self.pointer,
                        self._get_mouse_handler(current_index),
                    )
                )
                prompts.append(("", " ", self._get_mouse_handler(current_index)))
                prompts.append(
                    (
                        "class:selected",
                        choice.name.strip() + "\n",
                        self._get_mouse_handler(current_index),
                    )
                )
            else:
                prompts.append(
                    (
                        "",
                        " " * len(self.pointer),
                        self._get_mouse_handler(current_index),
                    )
                )
                prompts.append(("", " ", self._get_mouse_handler(current_index)))
                prompts.append(
                    (
                        "class:unselected",
                        choice.name.strip() + "\n",
                        self._get_mouse_handler(current_index),
                    )
                )
        return prompts

    def _get_error_prompt(self) -> AnyFormattedText:
        return self.error_message and [("class:error", self.error_message)]

    @lru_cache
    def _get_mouse_handler(
        self, index: Optional[int] = None
    ) -> Callable[[MouseEvent], None]:
        def _handle_mouse(event: MouseEvent) -> None:
            if event.event_type == MouseEventType.MOUSE_UP and index is not None:
                self._jump_to(index)
                if (
                    time() - self._last_mouse_up
                ) < 0.3 and index == self._last_mouse_index:
                    self._finish()
                else:
                    self._last_mouse_up = time()
                    self._last_mouse_index = index
            elif event.event_type == MouseEventType.SCROLL_UP:
                self._handle_up()
            elif event.event_type == MouseEventType.SCROLL_DOWN:
                self._handle_down()

        return _handle_mouse

    def _handle_up(self) -> None:
        if self.filtered_choices:
            self._jump_to((self._index - 1) % len(self.filtered_choices))

    def _handle_down(self) -> None:
        if self.filtered_choices:
            self._jump_to((self._index + 1) % len(self.filtered_choices))

    def _jump_to(self, index: int) -> None:
        self._index = index
        choice_num = len(self.filtered_choices)
        end_index = self._display_index + self.max_height - 2
        if self._index == self._display_index and self._display_index > 0:
            self._display_index -= 1
        elif self._index == choice_num - 1:
            start_index = choice_num - self.max_height + 1
            self._display_index = max(start_index, 0)
        elif self._index == end_index and end_index < choice_num - 1:
            self._display_index += 1
        elif self._index == 0:
            self._display_index = 0
